/**
 * Copyright (c) Microsoft Corporation
 *  All Rights Reserved
 *  MIT License
 *
* Permission is hereby granted, free of charge, to any person obtaining a copy of this 
* software and associated documentation files (the "Software"), to deal in the Software 
* without restriction, including without limitation the rights to use, copy, modify, 
* merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
* permit persons to whom the Software is furnished to do so, subject to the following 
* conditions:
*
* The above copyright notice and this permission notice shall be 
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS 
* OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
* WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT 
* OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

'use strict';

var types = {};

var Validator = function (config) {
  this.config = config;
};

Validator.prototype.validate = function (options) {

  var item,
    type,
    checker;

  if (!options) {
    options = {};
  }

  for (item in this.config) {
    if (this.config.hasOwnProperty(item)) {
      type = this.config[item];
      if (!type) {
        continue; // no need to validate
      }
      checker = types[type];
      if (!checker) { // missing required checker
        throw new TypeError('No handler to validate type ' + type + ' for item ' + item);
      }

      if (!checker.validate(options[item])) {
        throw new TypeError('Invalid value for ' + item + '. ' + checker.error);
      }
    } else {
      throw new TypeError('Missing value for ' + item);

    }
  }
};



Validator.isNonEmpty = 'isNonEmpty';
types.isNonEmpty = {
  validate: function (value) {
    return value !== '' && value !== undefined && value !== null;
  },
  error: 'The value cannot be empty'
};

Validator.isTypeLegal = 'isTypeLegal';
types.isTypeLegal = {
  validate: function (value) {
    return value === 'id_token' || value === 'id_token code' || value === 'code';
  },
  error: 'The responseType: must be either id_token, id_token code, or code.'
};

Validator.isModeLegal = 'isModeLegal';
types.isModeLegal = {
  validate: function (value) {
    return value === 'query' || value === 'form_post';
  },
  error: 'The responseMode: must be either query or form_post.'
};

Validator.isURL = 'isURL';
types.isURL = {

  validate: function (value) {
    var pattern = new RegExp('^(https?:\\/\\/)?' + // protocol
      '((([a-z\\d]([a-z\\d-]*[a-z\\d])*)\\.)+[a-z]{2,}|' + // domain name
      '((\\d{1,3}\\.){3}\\d{1,3}))' + // OR ip (v4) address
      '(\\:\\d+)?(\\/[-a-z\\d%_.~+]*)*' + // port and path
      '(\\?[;&a-z\\d%_.~+=-]*)?' + // query string
      '(\\#[-a-z\\d_]*)?$', 'i'); // fragment locator
    if (!pattern.test(value)) {
      return false;
    } else {
      return true;
    }
  },
  error: 'The URL must be valid and be https://'
};

exports.Validator = Validator;
